///////////////////////////////////////////////////////////////////////
// sysex.cpp :	Implements the sys_EX class
//	

#include <stdafx.h>
#include <windowsx.h>
#include <tchar.h>
#include <stdio.h>

#ifdef _UNICODE
#ifndef UNICODE
#define UNICODE
#endif
#endif

#if defined(_INC_COMDEF) || defined(SYSEX_INC_COMDEF)
#include <crtdbg.h>
#endif 

#include "sysex.h"

static TCHAR szEOS=_T('\0');

///////////////////////////////////////////////////////////////////////
// Helper functions forward prototyes

LPTSTR GetModuleName(LPTSTR pszModule);
LPTSTR RemoveFilenamePath(LPTSTR pszPath);

///////////////////////////////////////////////////////////////////////
// External variables

HINSTANCE g_hMODInstance=NULL;
LPCTSTR g_pszAppException=_T("Internal implementation error.");

///////////////////////////////////////////////////////////////////////
// class sys_EX

sys_EX::sys_EX(int nLine,char* pszFile) :	m_dwError(0L),
														m_pszSource(NULL),
														m_pszDescription(NULL)
{
	TCHAR szSource[_MAX_PATH];
	TCHAR szModule[_MAX_PATH];	

#ifdef UNICODE
	mbstowcs(szSource,pszFile,sizeof(szSource));
#else
	lstrcpyn(szSource,pszFile,sizeof(szSource));
#endif
	
	TCHAR szTemp[_MAX_PATH+_MAX_PATH+1];

	_stprintf(szTemp,_T("[module] %s [file] %s [line] %d"),
						  _tcslwr(GetModuleName(szModule)),
						  _tcslwr(RemoveFilenamePath(szSource)),
						  nLine);

	if((m_pszSource=new TCHAR[lstrlen(szTemp)+1])!=NULL)
		lstrcpy(m_pszSource,szTemp);
}

sys_EX::sys_EX(const sys_EX& o) : m_dwError(0L),
											 m_pszSource(NULL),
											 m_pszDescription(NULL)
{
	*this=o;
}

sys_EX::~sys_EX() 
{ 
	if(m_pszSource)
		delete m_pszSource;
	if(m_pszDescription)
		delete m_pszDescription;
}

DWORD sys_EX::GetError()
{
	return m_dwError;
}

LPCTSTR sys_EX::GetSource()
{
	return(m_pszSource?m_pszSource:_T(""));
}

LPCTSTR sys_EX::GetErrorString()
{
	if(!FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM|
						   FORMAT_MESSAGE_IGNORE_INSERTS,
						   NULL,m_dwError,MAKELANGID(LANG_NEUTRAL,SUBLANG_DEFAULT), 
						   m_szSYSError,sizeof(m_szSYSError),NULL))
	{
		m_szSYSError[0]=szEOS;
	}
	else
	{
		int nLen=lstrlen(m_szSYSError)-1;

		while(nLen>0 && (m_szSYSError[nLen]==_T('\n') || m_szSYSError[nLen]==_T('\r')))
			m_szSYSError[nLen--]=szEOS;	
	}

	return m_szSYSError;	
}

LPCTSTR sys_EX::GetDescription()
{
	return(m_pszDescription?m_pszDescription:_T(""));
}

void sys_EX::SetError(DWORD dwError)
{
	m_dwError=dwError;
}

void sys_EX::SetDescription(LPCTSTR pszFmt,...)
{
	if(!pszFmt || m_pszDescription)
		return;
	
	va_list argList;
	va_start(argList,pszFmt);

	TCHAR szTemp[1024];
	_vstprintf(szTemp,pszFmt,argList);
	va_end(argList);

	if((m_pszDescription=new TCHAR[lstrlen(szTemp)+1])!=NULL)
		lstrcpy(m_pszDescription,szTemp);
}

sys_EX& sys_EX::operator=(const sys_EX& o)
{
	if(this!=&o)
	{
		m_dwError=o.m_dwError;

		if(m_pszSource)
			delete m_pszSource;
		m_pszSource=NULL;

		if(o.m_pszSource)
			if((m_pszSource=new TCHAR[lstrlen(o.m_pszSource)+1])!=NULL)
				lstrcpy(m_pszSource,o.m_pszSource);

		if(m_pszDescription)
			delete m_pszDescription;
		m_pszDescription=NULL;

		if(o.m_pszDescription)
			if((m_pszDescription=new TCHAR[lstrlen(o.m_pszDescription)+1])!=NULL)
				lstrcpy(m_pszDescription,o.m_pszDescription);
	}

	return *this;
}

#if !defined(_INC_COMDEF) && !defined(SYSEX_INC_COMDEF)
#pragma message ("Define SYSEX_INC_COMDEF in your project to link in _com_error support.")
#else

HRESULT sys_EX::raise_com_error(REFGUID rguid)
{
	HRESULT hr;

	_bstr_t bstrSource=GetSource();
	BSTR BSTRSource=SysAllocString(bstrSource);

	TCHAR szError[64];
	wsprintf(szError,_T("%d|"),m_dwError);
	_bstr_t bstrDescription=szError;
	bstrDescription+=GetDescription();
	BSTR BSTRDescription=SysAllocString(bstrDescription);

	IErrorInfo *perrinfo;
	ICreateErrorInfo *pcerrinfo;

	hr=CreateErrorInfo(&pcerrinfo);
	_ASSERTE(SUCCEEDED(hr));
		
	pcerrinfo->SetGUID(rguid);
	pcerrinfo->SetSource(BSTRSource);
	pcerrinfo->SetDescription(BSTRDescription);

	hr=pcerrinfo->QueryInterface(IID_IErrorInfo,(LPVOID FAR*)&perrinfo);
	_ASSERTE(SUCCEEDED(hr));

	hr=SetErrorInfo(0,perrinfo);
	_ASSERTE(SUCCEEDED(hr));
	
	SysFreeString(BSTRSource);
	SysFreeString(BSTRDescription);
	perrinfo->Release();
	pcerrinfo->Release();

	return SYSEX_E_COMERROR;
}

void sys_EX::set_com_error(_com_error& e)
{
	if((LPCTSTR)e.Source())
	{
		if(!m_pszSource)
		{
			if((m_pszSource=new TCHAR[lstrlen((LPCTSTR)e.Source())+1])!=NULL)
				lstrcpy(m_pszSource,(LPCTSTR)e.Source());
		}
		else
		{
			LPTSTR psz=new TCHAR[lstrlen(m_pszSource)+lstrlen((LPCTSTR)e.Source())+3];
			if(psz)
			{
				lstrcpy(psz,(LPCTSTR)e.Source());
				lstrcat(psz,_T("\r\n"));
				lstrcat(psz,m_pszSource);

				delete m_pszSource;
				if((m_pszSource=new TCHAR[lstrlen(psz)+1])!=NULL)
					lstrcpy(m_pszSource,psz);
				delete psz;
			}
		}
	}

	if(m_pszDescription)
		delete m_pszDescription;
	m_pszDescription=NULL;

	if(e.Error()!=SYSEX_E_COMERROR)
	{
		m_dwError=e.Error();

		if((LPCTSTR)e.Description())
			if((m_pszDescription=new TCHAR[lstrlen((LPCTSTR)e.Description())+1])!=NULL)
				lstrcpy(m_pszDescription,(LPCTSTR)e.Description());
	}
	else
	{
		if((LPCTSTR)e.Description())
		{
			TCHAR szTemp[2048];
			int nRead=_stscanf((LPCTSTR)e.Description(),_T("%d|%s"),&m_dwError,szTemp);
			_ASSERTE(nRead>0 && nRead<=2);

			if((m_pszDescription=new TCHAR[lstrlen(szTemp)+1])!=NULL)
				lstrcpy(m_pszDescription,szTemp);
		}
	}
}

#endif // !defined(_INC_COMDEF) && !defined(SYSEX_INC_COMDEF)

///////////////////////////////////////////////////////////////////////
// Dialog construction and display

#define CNTL_BUTTON_OK					0
#define CNTL_DESCRIPTION_STATIC		1
#define CNTL_SYSTEM_ERROR_STATIC		2
#define CNTL_SOURCE_STATIC				3
#define CNTL_DESCRIPTION_EDIT			4
#define CNTL_SYSTEM_ERROR_EDIT		5
#define CNTL_SOURCE_EDIT				6
#define CNTL_MAX							7

struct sys_EXCTRL
{
	DLGITEMTEMPLATE m_Item;
	WORD				 m_nClassIDHI;
	WORD				 m_nClassIDLO;
	LPCTSTR			 m_pszText;
};

#pragma warning ( disable : 4245 )

static sys_EXCTRL g_dlgControl[CNTL_MAX]=
{
	{ {BS_PUSHBUTTON|WS_VISIBLE|WS_CHILD,0L,168,97,50,14,IDOK}, 0xFFFF,0X0080,_T("Close") },
	{ {WS_VISIBLE|WS_CHILD,0L,5,8,37,8,-1}, 0xFFFF,0X0082,_T("Description") },
	{ {WS_VISIBLE|WS_CHILD,0L,5,40,25,8,-1}, 0xFFFF,0X0082,_T("System") },
	{ {WS_VISIBLE|WS_CHILD,0L,5,63,24,8,-1}, 0xFFFF,0X0082,_T("Source") },
	{ {WS_VISIBLE|WS_CHILD|WS_BORDER|ES_MULTILINE|ES_AUTOVSCROLL|ES_READONLY,0L,43,7,182,28,1000}, 0xFFFF,0X0081,_T("") },
	{ {WS_VISIBLE|WS_CHILD|WS_BORDER|ES_MULTILINE|ES_AUTOVSCROLL|ES_READONLY,0L,43,38,182,21,1001}, 0xFFFF,0X0081,_T("") },
	{ {WS_VISIBLE|WS_CHILD|WS_BORDER|ES_MULTILINE|ES_AUTOVSCROLL|ES_READONLY,0L,43,62,182,28,1002}, 0xFFFF,0X0081,_T("") }
};

#pragma warning ( default : 4245 )

BOOL CALLBACK MessageProc(HWND hWnd,UINT nMsg,WPARAM wParam,LPARAM)
{
	switch(nMsg)
	{
		case WM_INITDIALOG:
			SetWindowPos(hWnd,HWND_TOPMOST,0,0,0,0,SWP_NOSIZE|SWP_NOMOVE);
			return TRUE;

		case WM_COMMAND:
			if(LOWORD(wParam)==IDOK || LOWORD(wParam)==IDCANCEL)
			{
				DestroyWindow(hWnd);
				return TRUE;
			}
			
		case WM_SYSCOMMAND:
			if(wParam==SC_CLOSE)
			{
				DestroyWindow(hWnd);
				return TRUE;
			}	
	}

	return FALSE;
}

void SetText(PWORD& pwData,LPCTSTR pszText)
{
	int nLen=lstrlen(pszText);

#ifndef UNICODE
	nLen=MultiByteToWideChar(GetACP(),MB_PRECOMPOSED,pszText,nLen,pwData,nLen)+1;
#else
	lstrcpy(pwData,pszText);
#endif

	pwData+=nLen;	
}

PWORD Align(PWORD pwData)
{
  ULONG n=(ULONG)pwData;

  n+=3;
  n>>=2;
  n<<=2;

  return (PWORD)n;
}

void PopulateTemplateItems(PWORD& pwData)
{
	for(int n=0; n<CNTL_MAX; n++)
	{
		pwData=Align(pwData);

		*pwData++=LOWORD(g_dlgControl[n].m_Item.style);
		*pwData++=HIWORD(g_dlgControl[n].m_Item.style);
		*pwData++=0;         
		*pwData++=0;       
		*pwData++=g_dlgControl[n].m_Item.x;      
		*pwData++=g_dlgControl[n].m_Item.y;      
		*pwData++=g_dlgControl[n].m_Item.cx;     
		*pwData++=g_dlgControl[n].m_Item.cy;     
		*pwData++=g_dlgControl[n].m_Item.id;    
		*pwData++=g_dlgControl[n].m_nClassIDHI;       
		*pwData++=g_dlgControl[n].m_nClassIDLO;      

		SetText(pwData,g_dlgControl[n].m_pszText);

		*pwData++=0;
	}
}

void PopulateTemplate(PWORD& pwData,LPCTSTR pszTitle)
{
	DWORD dwStyle=WS_POPUP|WS_CAPTION|WS_VISIBLE|
					  DS_MODALFRAME|DS_SETFONT|DS_CENTER;

  *pwData++=LOWORD(dwStyle);
  *pwData++=HIWORD(dwStyle);
  *pwData++=0;
  *pwData++=0;
  *pwData++=CNTL_MAX;
  *pwData++=0;
  *pwData++=0;
  *pwData++=234;
  *pwData++=118;
  *pwData++=0;
  *pwData++=0;

	SetText(pwData,pszTitle);
	
	*pwData++=8;

	SetText(pwData,_T("MS Sans Serif"));

	PopulateTemplateItems(pwData);
}

void sys_EX::Display(HWND hWndParent) 
{ 	
	if(!hWndParent || !IsWindow(hWndParent))
		hWndParent=GetDesktopWindow();

	int nAlloc=0;

	g_dlgControl[CNTL_DESCRIPTION_EDIT].m_pszText=GetDescription();

	if(GetError())
		g_dlgControl[CNTL_SYSTEM_ERROR_EDIT].m_pszText=GetErrorString();
	else
		g_dlgControl[CNTL_SYSTEM_ERROR_EDIT].m_pszText=_T("");
		
	g_dlgControl[CNTL_SOURCE_EDIT].m_pszText=GetSource();

	for(int n=0; n<CNTL_MAX; n++)
		nAlloc+=sizeof(TCHAR)*(lstrlen(g_dlgControl[n].m_pszText)+1);
	
	nAlloc+=sizeof(DLGTEMPLATE)+
			  sizeof(sys_EXCTRL)*CNTL_MAX+
			  sizeof(DWORD)*CNTL_MAX;

	PWORD pwData=NULL;
	PWORD pDlgTemplate=NULL;

	pwData=pDlgTemplate=(PWORD)HeapAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,nAlloc*2);

	if(!pDlgTemplate)
		return;

	TCHAR szModule[_MAX_PATH];	
	PopulateTemplate(pwData,GetModuleName(szModule));

	HWND hWnd=CreateDialogIndirect(GetModuleHandle(NULL),(LPDLGTEMPLATE)pDlgTemplate,hWndParent,(DLGPROC)MessageProc);

	HeapFree(GetProcessHeap(),0,pDlgTemplate);

	MSG msg;
	while(IsWindow(hWnd) && GetMessage(&msg,NULL,0,0))
	{
		if(!IsDialogMessage(hWnd,&msg))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}
}

///////////////////////////////////////////////////////////////////////
// Helper functions

LPTSTR GetModuleName(LPTSTR pszModule)
{  
	if(g_hMODInstance==NULL)
		g_hMODInstance=GetModuleHandle(NULL);

	if(!GetModuleFileName(g_hMODInstance,pszModule,_MAX_PATH))
		return _T("");

	TCHAR szSource[_MAX_PATH];
	TCHAR szDrive[_MAX_DRIVE];
	TCHAR szDir[_MAX_DIR];
	TCHAR szFilename[_MAX_FNAME];
	TCHAR szExt[_MAX_EXT];

	lstrcpyn(szSource,pszModule,sizeof(szSource));

	_tsplitpath(szSource,szDrive,szDir,szFilename,szExt); 

	lstrcpy(pszModule,szFilename);
	
	return pszModule;
}

LPTSTR RemoveFilenamePath(LPTSTR pszPath)
{  
	TCHAR szSource[_MAX_PATH];
	TCHAR szDrive[_MAX_DRIVE];
	TCHAR szDir[_MAX_DIR];
	TCHAR szFilename[_MAX_FNAME];
	TCHAR szExt[_MAX_EXT];

	lstrcpyn(szSource,pszPath,sizeof(szSource));

	_tsplitpath(szSource,szDrive,szDir,szFilename,szExt); 

	lstrcpy(szSource,szFilename);
	lstrcat(szSource,szExt);

	lstrcpy(pszPath,szSource);
	
	return pszPath;
}
